#ifndef PARTICLETRACKER_UTILIIES_HH
#define PARTICLETRACKER_UTILIIES_HH

#include "Vector.hh"
#include "grid.hh"
#include "gw_bc.hh"

// Default values for the precompiler macros

// default is no gravity
#ifndef GRAVITY_AXES
#define GRAVITY_AXES -1
#endif



class ConstantDispersion
{
private:
  double default_value_;
  
public:
  ConstantDispersion(double _default_value)
    : default_value_(_default_value)
  {}
  
  double operator()(Vector ) const 
  {
    return default_value_;
  }
};




class PressureVelocityConverter
{
private:
  const Vector & pressure_;
  Vector cell_length;
  const Grid & grid_;
  const std::vector<size_t> cells;
  const size_t grid_size;
  std::vector<Vector> flux_field;
  const ConductivityCoefficients & conductivity_;
  const BoundaryConditions & boundary_conditions_;
  const double porosity_;
  const size_t X,XY;
  
public:
  PressureVelocityConverter( const Vector & _pressure,
							 const BoundaryConditions & _pressure_boundary_conditions,
							 const Grid & grid,
							 const ConductivityCoefficients & _conductivity,
							 const double _porosity
							 )
	:
    pressure_(_pressure),
    cell_length(Grid::dimensions),
    grid_(grid), 
    cells(grid.cells()),
    grid_size(std::accumulate(cells.begin(),cells.end(),1,std::multiplies<size_t>())),
    flux_field(grid_size),
    conductivity_(_conductivity),
    boundary_conditions_(_pressure_boundary_conditions),
    porosity_(_porosity),
    X(grid.cells()[0]),
    XY(grid.cells()[0]*grid.cells()[1])

  {
    for( ElementIterator it = grid_.begin()
		   ; it != grid_.end()
		   ; ++it
		 )
	  {
		
		const size_t e_id = it.id();
		flux_field[e_id].resize( Grid::dimensions*2 );
		
		for( FaceIterator fit  = (*it).begin()
			   ; fit!= (*it).end()
			   ; ++fit)
		  {

			double flux(0.0);
			
			const unsigned int normal_component = (*fit).normal_component();
			const int normal_direction = (*fit).normal_direction();

			if( !(*fit).is_boundary() )
			  {
				const size_t f_id = fit.out_id();

				const double k_e = conductivity_(e_id);
				const double k_f = conductivity_(f_id);
				const double k = 2. / (1./k_e + 1./k_f);

				const double J = - (double)normal_direction * 
				  (pressure_[f_id]-pressure_[e_id])
				  / (*fit).dist_to_barycenter() *  0.5 * k;

				flux +=  J;

				if(normal_component==(unsigned int)GRAVITY_AXES){
				  const double G = k;
				  flux += - G;
				}
			  }
			else
			  {
				Vector barycenter = (*it).barycenter();
				const double h_2 = (*it).extent()[normal_component] * 0.5;
				barycenter[normal_component] += (double)normal_direction * h_2;
				
				// if dirichlet condition
				double dirichlet_value=0;
				if(boundary_conditions_.is_dirichlet(barycenter, &dirichlet_value))
				  {
					const double k_e = conductivity_(e_id);
					const double k = k_e;
					
					const double J = - (double)normal_direction * (dirichlet_value - pressure_[e_id])
					  / h_2 * k;
					
					flux += J;
				  }
				// if neumann condition
				else
				  {
					double neumann_value = boundary_conditions_.get_neumann(barycenter);
					flux += neumann_value;
				  }
			  }
			
			const unsigned int inc = (normal_direction > 0) ? 1 : 0;
			flux_field[ e_id ][ normal_component*2 + inc ] = flux / porosity_;
		  }//fit
		
	  }//it
	
    for(unsigned int d=0; d < Grid::dimensions; ++d)
      cell_length[d] = grid.extent()[d] / (double)grid_.cells()[d];
	
  }
  
  Vector operator()(Vector position) const
  {
    std::vector<size_t> x(Grid::dimensions);
    for(int d=0; d<Grid::dimensions; ++d)
	  {
		// compute integer grid coordinates
		x[d] = (size_t)( position[d]/cell_length[d] );
		
		// compute local coordinates
		position[d] -= cell_length[d] * (double)x[d];
		position[d] /= cell_length[d];
	  }
	
    const size_t z_coord( Grid::dimensions == 3 ? x[2] : 0 );
    const size_t index( x[0] + x[1]*X + z_coord * XY );
    
    Vector flux(Grid::dimensions,0.0);
    for(int d=0; d<Grid::dimensions; ++d)
	  {
		flux[d] = flux_field[ index ][ d*2 ] * ( 1.0 - position[ d ] )
		  + position[ d ] * flux_field[ index ][ d*2 + 1];
	  }
    return flux;
  }
  
};

#endif // PARTICLETRACKER_UTILIIES_HH
